@@ -2,6 +2,8 @@ use utf8;
 
 use Scalar::Util ();
 
+### GO CLOCK WIDGET #########################################################
+
 package game::goclock;
 
 # Lo and Behold! I admit it! The rounding stuff etc.. in goclock
@@ -34,7 +36,7 @@ sub configure {
 
    if ($timesys == TIMESYS_ABSOLUTE) {
       $self->{format} = sub {
-         if ($_[0] <= 0) {
+         if ($_[0] < 0) {
             "TIMEOUT";
          } else {
             util::format_time $_[0];
@@ -45,7 +47,7 @@ sub configure {
       my $low = $interval * $count;
 
       $self->{format} = sub {
-         if ($_[0] <= 0) {
+         if ($_[0] < 0) {
             "TIMEOUT";
          } elsif ($_[0] > $low) {
             util::format_time $_[0] - $low;
@@ -58,7 +60,7 @@ sub configure {
 
    } elsif ($timesys == TIMESYS_CANADIAN) {
       $self->{format} = sub {
-         if ($_[0] <= 0) {
+         if ($_[0] < 0) {
             "TIMEOUT";
          } elsif (!$self->{moves}) {
             util::format_time $_[0] - $low;
@@ -129,28 +131,44 @@ sub stop {
    remove Glib::Source delete $self->{timeout} if $self->{timeout};
 }
 
+### USER PANEL ##############################################################
+
 package game::userpanel;
 
+use KGS::Constants;
+
 use Glib::Object::Subclass
-   Gtk2::HBox,
+   Gtk2::Frame,
    properties => [
-      Glib::ParamSpec->IV ("colour", "colour", "User Colour", 0, 1, 0, [qw(construct-only writable)]),
+      Glib::ParamSpec->IV ("colour", "colour", "User Colour",
+                           COLOUR_BLACK, COLOUR_WHITE, COLOUR_BLACK,
+                           [qw(construct-only readable writable)]),
    ];
 
 sub INIT_INSTANCE {
    my ($self) = @_;
 
-   $self->add (my $vbox = new Gtk2::VBox);
+   $self->add ($self->{window} = my $window = new Gtk2::EventBox); # for bg
 
-   $vbox->add ($self->{name} = new Gtk2::Label $self->{name});
-   $vbox->add ($self->{info} = new Gtk2::Label "");
-   $vbox->add ($self->{clock} = new game::goclock); Scalar::Util::weaken $self->{clock};
+   $window->add (my $vbox = new Gtk2::VBox);
+
+   $vbox->pack_start (($self->{name} = new Gtk2::Label "-"), 1, 1, 0);
+   $vbox->pack_start (($self->{info} = new Gtk2::Label "-"), 1, 1, 0);
+   $vbox->pack_start (($self->{clock} = new game::goclock), 1, 1, 0);
 
    $vbox->add ($self->{imagebox} = new Gtk2::VBox);
 
    $self;
 }
 
+sub SET_PROPERTY {
+   my ($self, $pspec, $value) = @_;
+
+   $self->{$pspec->get_name} = $value;
+
+   $self->set_name ("userpanel-$self->{colour}");
+}
+
 sub configure {
    my ($self, $app, $user, $rules) = @_;
 
@@ -190,6 +208,8 @@ sub set_timer {
    $self->{clock}->set_time ($start, $time, $moves);
 }
 
+### GAME WINDOW #############################################################
+
 package game;
 
 use Scalar::Util qw(weaken);
@@ -210,10 +230,12 @@ use POSIX qw(ceil);
 
 sub new {
    my ($self, %arg) = @_;
+
    $self = $self->Glib::Object::new;
    $self->{$_} = delete $arg{$_} for keys %arg;
 
-   gtk::state $self, "game::window", undef, window_size => [600, 500];
+   gtk::state $self, "game::window", undef, window_size => [620, 460];
+   $self->set (allow_shrink => 1);
 
    $self->signal_connect (destroy => sub {
       $self->unlisten;
@@ -222,20 +244,16 @@ sub new {
    });#d#
 
    $self->add (my $hpane = new Gtk2::HPaned);
-   gtk::state $hpane, "game::hpane", undef, position => 500;
+   gtk::state $hpane, "game::hpane", undef, position => 420;
 
    # LEFT PANE
 
    $hpane->pack1 (($self->{left} = new Gtk2::VBox), 1, 0);
    
-   $self->{boardbox} = new Gtk2::VBox;
-
    $hpane->pack1((my $vbox = new Gtk2::VBox), 1, 1);
 
    # board box (aspect/canvas)
    
-   #$self->{boardbox}->pack_start((my $frame = new Gtk2::Frame), 0, 1, 0);
-
    # RIGHT PANE
 
    $hpane->pack2 ((my $vbox = new Gtk2::VBox), 1, 1);
@@ -245,11 +263,11 @@ sub new {
 
    {
       $frame->add (my $vbox = new Gtk2::VBox);
-      $vbox->add ($self->{title} = new Gtk2::Label $title);
+      $vbox->add ($self->{title} = new Gtk2::Label "-");
 
       $vbox->add (my $hbox = new Gtk2::HBox);
 
-      $hbox->pack_start (($self->{board_label} = new Gtk2::Label), 0, 1, 0);
+      $hbox->pack_start (($self->{board_label} = new Gtk2::Label), 0, 0, 0);
 
       $self->{moveadj} = new Gtk2::Adjustment 1, 1, 1, 1, 5, 0;
 
@@ -271,22 +289,25 @@ sub new {
    $vbox->pack_start ((my $buttonbox = new Gtk2::HButtonBox), 0, 1, 0);
 
    $buttonbox->add ($self->{button_pass} =
-      Gtk2::Button->Glib::Object::new (label => "Pass", no_show_all => 1, visible => 0));
+      Gtk2::Button->Glib::Object::new (label => "Pass", visible => 0));
    $self->{button_pass}->signal_connect (clicked => sub {
       $self->{board_click}->(255, 255) if $self->{board_click};
    });
+   eval { $self->{button_pass}->set (no_show_all => 1) }; # workaround for gtk+-2.2
    $buttonbox->add ($self->{button_undo} =
-      Gtk2::Button->Glib::Object::new (label => "Undo", no_show_all => 1, visible => 0));
+      Gtk2::Button->Glib::Object::new (label => "Undo", visible => 0));
    $self->{button_undo}->signal_connect (clicked => sub {
       $self->send (req_undo => channel => $self->{channel});
    });
+   eval { $self->{button_undo}->set (no_show_all => 1) }; # workaround for gtk+-2.2
    $buttonbox->add ($self->{button_resign} =
-      Gtk2::Button->Glib::Object::new (label => "Resign", no_show_all => 1, visible => 0));
+      Gtk2::Button->Glib::Object::new (label => "Resign", visible => 0));
    $self->{button_resign}->signal_connect (clicked => sub {
       $self->send (resign_game => channel => $self->{channel}, player => $self->{colour});
    });
+   eval { $self->{button_resign}->set (no_show_all => 1) }; # workaround for gtk+-2.2
    
-   $vbox->pack_start (($self->{chat} = new superchat), 1, 1, 0);
+   $vbox->pack_start (($self->{chat} = new chat), 1, 1, 0);
 
    $self->set_channel ($self->{channel});
 
@@ -318,30 +339,7 @@ sub set_channel {
    }
 }
 
-sub event_update_users {
-   my ($self, $add, $update, $remove) = @_;
-
-#   $self->{userlist}->update ($add, $update, $remove);
-
-   $self->{challenge}{$_->{name}} && (delete $self->{challenge}{$_->{name}})->{inlay}->destroy
-      for @$remove;
-
-   $self->{users_inlay}->refresh;
-
-   my %important;
-   $important{$self->{black}{name}}++;
-   $important{$self->{white}{name}}++;
-   $important{$self->{owner}{name}}++;
-
-   if (my @users = grep $important{$_->{name}}, @$add) {
-      $self->{chat}->append_text ("\n<header>Joins:</header>");
-      $self->{chat}->append_text (" <user>" . $_->as_string . "</user>") for @users;
-   }
-   if (my @users = grep $important{$_->{name}}, @$remove) {
-      $self->{chat}->append_text ("\n<header>Parts:</header>");
-      $self->{chat}->append_text (" <user>" . $_->as_string . "</user>") for @users;
-   }
-}
+### JOIN/LEAVE ##############################################################
 
 sub join {
    my ($self) = @_;
@@ -350,210 +348,11 @@ sub join {
    $self->SUPER::join;
 }
 
-sub update_cursor {
+sub part {
    my ($self) = @_;
 
-   my $running = $self->{showmove} == @{$self->{path}} && $self->is_active;
-
-   delete $self->{board_click};
-
-   if ($self->{teacher} eq $self->{app}{conn}) {
-      #TODO# # teaching mode not implemented
-      $self->{button_pass}->set (label => "Pass", visible => 1, sensitive => 1);
-      $self->{button_undo}->hide;
-      $self->{button_resign}->hide;
-      $self->{board}->set (cursor => undef);
-
-   } elsif ($running && $self->{colour} != COLOUR_NONE) {
-      # during game
-      $self->{button_undo}->show;
-      $self->{button_resign}->show;
-
-      if ($self->{cur_board}{score}) {
-         # during scoring
-         $self->{button_pass}->set (label => "Done", visible => 1, sensitive => 1);
-         $self->{board}->set (cursor => sub {
-            $_[0] & (MARK_B | MARK_W)
-               ? $_[0] ^ MARK_GRAYED
-               : $_[0];
-         });
-         $self->{board_click} = sub {
-            if ($_[0] == 255) {
-               $self->{button_pass}->sensitive (0);
-               $self->done;
-            } else {
-               $self->send (mark_dead =>
-                  channel => $self->{channel},
-                  x       => $_[0],
-                  y       => $_[1],
-                  dead    => !($self->{cur_board}{board}[$_[0]][$_[1]] & MARK_GRAYED),
-               );
-            }
-         };
-
-      } elsif ($self->{colour} == $self->{whosemove}) {
-         # normal move
-         $self->{button_pass}->set (label => "Pass", visible => 1, sensitive => 1);
-         $self->{board}->set (cursor => sub {
-            # if is_valid_move oder so#TODO#
-            $_[0] & (MARK_B | MARK_W)
-               ? $_[0]
-               : $_[0] | MARK_GRAYED | ($self->{colour} == COLOUR_WHITE ? MARK_W : MARK_B);
-         });
-         $self->{board_click} = sub {
-            $self->send (game_move => channel => $self->{channel}, x => $_[0], y => $_[1]);
-            $self->{board}->set (cursor => undef);
-            delete $self->{board_click};
-            $self->{button_pass}->sensitive (0);
-         };
-      } else {
-         $self->{button_pass}->set (label => "Pass", sensitive => 0, visible => 1);
-         $self->{board}->set (cursor => undef);
-      }
-   } else {
-      $self->{button_undo}->hide;
-      $self->{button_resign}->hide;
-      $self->{button_pass}->hide;
-      $self->{board}->set (cursor => undef);
-      #TODO# # implement coordinate-grabbing
-   }
-}
-
-sub update_timers {
-   my ($self, $timers) = @_;
-
-   my $running = $self->{showmove} == @{$self->{path}} && !$self->{teacher};
-
-   for my $colour (COLOUR_BLACK, COLOUR_WHITE) {
-      my $t = $timers->[$colour];
-      $self->{userpanel}[$colour]->set_timer (
-            $running && $colour == $self->{whosemove} && $t->[0],
-            $t->[1] || $self->{rules}{time}
-                       + ($self->{rules}{timesys} == TIMESYS_BYO_YOMI
-                         && $self->{rules}{interval} * $self->{rules}{count}),
-            $t->[2] || $self->{rules}{count});
-   }
-}
-
-sub update_board {
-   my ($self) = @_;
-   return unless $self->{path};
-
-   $self->{board_label}->set_text ("Move " . ($self->{showmove} - 1));
-
-   $self->{cur_board} = new KGS::Game::Board $self->{size};
-   $self->{cur_board}->interpret_path ([@{$self->{path}}[0 .. $self->{showmove} - 1]]);
-
-   $self->{userpanel}[$_]->set_captures ($self->{cur_board}{captures}[$_])
-      for COLOUR_WHITE, COLOUR_BLACK;
-
-   if ($self->{rules}{ruleset} == RULESET_JAPANESE) {
-      if ($self->{curnode}{move} == 0) {
-         $self->{whosemove} = $self->{handicap} ? COLOUR_WHITE : COLOUR_BLACK;
-      } else {
-         $self->{whosemove} = 1 - $self->{cur_board}{last};
-      }
-   } else {
-      # Chinese, Aga, NZ all have manual placement
-      if ($self->{curnode}{move} < $self->{handicap}) {
-         $self->{whosemove} = COLOUR_BLACK;
-      } elsif ($self->{curnode}{move} == $self->{handicap}) {
-         $self->{whosemove} = $self->{handicap} ? COLOUR_WHITE : COLOUR_BLACK;
-      } else {
-         $self->{whosemove} = 1 - $self->{cur_board}{last};
-      }
-   }
-
-   my $start_time = $self->{rules}{time};
-
-   if ($self->{showmove} == @{$self->{path}}) {
-      $self->{timers} = [
-         [$self->{lastmove_time}, @{$self->{cur_board}{timer}[0]}],
-         [$self->{lastmove_time}, @{$self->{cur_board}{timer}[1]}],
-      ];
-      $self->update_timers ($self->{timers});
-   } else {
-      $self->update_timers ([
-         [0, @{$self->{cur_board}{timer}[0]}],
-         [0, @{$self->{cur_board}{timer}[1]}],
-      ]);
-   }
-
-   $self->{board}->set_board ($self->{cur_board});
-
-   if ($self->{cur_board}{score}) {
-      $self->{score_inlay} ||= $self->{chat}->new_inlay;
-      $self->{score_inlay}->clear;
-      $self->{score_inlay}->append_text ("\n<header>Scoring</header>"
-                                         . "\n<score>"
-                                         . "White: $self->{cur_board}{score}[COLOUR_WHITE], "
-                                         . "Black: $self->{cur_board}{score}[COLOUR_BLACK]"
-                                         . "</score>");
-   } elsif ($self->{score_inlay}) {
-      (delete $self->{score_inlay})->clear;
-   }
-
-   $self->update_cursor;
-}
-
-sub event_update_tree {
-   my ($self) = @_;
-
-   (delete $self->{undo_inlay})->clear
-      if $self->{undo_inlay};
-
-   $self->{path} = $self->get_path;
-
-   if ($self->{moveadj}) {
-      my $upper = $self->{moveadj}->upper;
-      my $pos = $self->{moveadj}->get_value;
-      my $move = scalar @{$self->{path}};
-
-      $self->{moveadj}->upper ($move);
-      
-      $self->{moveadj}->changed;
-      if ($pos == $upper) {
-         $self->{moveadj}->value ($move);
-         $self->{moveadj}->value_changed;
-      }
-   }
-}
-
-sub event_update_comments {
-   my ($self, $node, $comment, $newnode) = @_;
-   $self->SUPER::event_update_comments($node, $comment, $newnode);
-
-   my $text;
-
-   $text .= "\n<header>Move <move>$node->{move}</move>, Node <node>$node->{id}</node></header>"
-      if $newnode;
-
-   for (split /\n/, $comment) {
-      $text .= "\n";
-      if (s/^([0-9a-zA-Z]+ \[[0-9dkp\?\-]+\])://) {
-         $text .= "<user>" . (util::toxml $1) . "</user>:";
-      }
-      
-      # coords only for 19x19 so far
-      $_ = util::toxml $_;
-      s{
-         (
-            \b
-            (?:[bw])?
-            [, ]{0,2}
-            [a-hj-t] # valid for upto 19x19
-            \s?
-            [1-9]?[0-9]
-            \b
-         )
-      }{
-         "<coord>$1</coord>";
-      }sgexi;
-
-      $text .= $_;
-   }
-
-   $self->{chat}->append_text ($text);
+   $self->hide;
+   $self->SUPER::part;
 }
 
 sub event_join {
@@ -571,53 +370,50 @@ sub event_part {
    $self->destroy;
 }
 
-sub event_move {
-   my ($self, $pass) = @_;
-
-   sound::play 1, $pass ? "pass" : "move";
-}
-
-sub event_update_game {
+sub event_quit {
    my ($self) = @_;
 
-   $self->SUPER::event_update_game;
-
-   return unless $self->{joined};
-
-   $self->{colour} = $self->player_colour ($self->{conn}{name});
-   
-   my $title = defined $self->{channel}
-                  ? $self->owner->as_string . " " . $self->opponent_string
-                  : "Game Window";
-   $self->set_title ("KGS Game $title");
-   $self->{title}->set_text ($title);
-
-   $self->{user}[COLOUR_BLACK] = $self->{black};
-   $self->{user}[COLOUR_WHITE] = $self->{white};
-
-   # show board
-   if ($self->is_inprogress) {
-      if (!$self->{boardbox}->parent) {
-         $self->{boardbox}->add ($self->{board} = new Gtk2::GoBoard size => $self->{size});
-         $self->{left}->add ($self->{boardbox});
-         $self->{board}->signal_connect (button_release => sub {
-            if ($_[1] == 1) {
-               $self->{board_click}->($_[2], $_[3]) if $self->{board_click};
-            }
-         });
-      }
-      if (my $ch = delete $self->{challenge}) {
-         $_->{inlay}->destroy for values %$ch;
-      }
-      $self->update_cursor;
-   }
-
-   $self->{left}->show_all;
-
-   $self->{rules_inlay}->refresh;
-   
+   $self->SUPER::event_quit;
+   $self->destroy;
 }
 
+### USERS ###################################################################
+
+sub draw_users {
+   my ($self, $inlay) = @_;
+
+   for (sort keys %{$self->{users}}) {
+      $inlay->append_text ("  <user>" . $self->{users}{$_}->as_string . "</user>");
+   }
+}
+
+sub event_update_users {
+   my ($self, $add, $update, $remove) = @_;
+
+#   $self->{userlist}->update ($add, $update, $remove);
+
+   $self->{challenge}{$_->{name}} && (delete $self->{challenge}{$_->{name}})->{inlay}->destroy
+      for @$remove;
+
+   $self->{users_inlay}->refresh;
+
+   my %important;
+   $important{$self->{black}{name}}++;
+   $important{$self->{white}{name}}++;
+   $important{$self->{owner}{name}}++;
+
+   if (my @users = grep $important{$_->{name}}, @$add) {
+      $self->{chat}->append_text ("\n<leader>Joins:</leader>");
+      $self->{chat}->append_text (" <user>" . $_->as_string . "</user>") for @users;
+   }
+   if (my @users = grep $important{$_->{name}}, @$remove) {
+      $self->{chat}->append_text ("\n<leader>Parts:</leader>");
+      $self->{chat}->append_text (" <user>" . $_->as_string . "</user>") for @users;
+   }
+}
+
+### GAME INFO ###############################################################
+
 sub draw_setup {
    my ($self, $inlay) = @_;
 
@@ -677,12 +473,64 @@ sub draw_setup {
 
 }
 
+sub event_update_game {
+   my ($self) = @_;
+
+   $self->SUPER::event_update_game;
+
+   return unless $self->{joined};
+
+   $self->{colour} = $self->player_colour ($self->{conn}{name});
+   
+   $self->{user}[COLOUR_BLACK] = $self->{black};
+   $self->{user}[COLOUR_WHITE] = $self->{white};
+
+   # show board
+   if ($self->is_inprogress) {
+      if (!$self->{board}) {
+         $self->{left}->add ($self->{board} = new Gtk2::GoBoard size => $self->{size});
+         $self->{board}->signal_connect (button_release => sub {
+            return unless $self->{cur_board};
+            if ($_[1] == 1) {
+               $self->{board_click}->($_[2], $_[3]) if $self->{board_click};
+            }
+         });
+         $self->{board}->show_all;
+      }
+      if (my $ch = delete $self->{challenge}) {
+         $_->{inlay}->destroy for values %$ch;
+      }
+      $self->update_cursor;
+   }
+
+   my $title = defined $self->{channel}
+                  ? $self->owner->as_string . " " . $self->opponent_string
+                  : "Game Window";
+   $self->set_title ("KGS Game $title");
+   $self->{title}->set_text ($title); # title gets redrawn wrongly
+
+   $self->{rules_inlay}->refresh;
+
+   if (exists $self->{teacher}) {
+      $self->{teacher_inlay} ||= $self->{chat}->new_inlay;
+      $self->{teacher_inlay}->clear;
+      $self->{teacher_inlay}->append_text ("\n<header>Teacher:</header> <user>"
+                                           . (util::toxml $self->{teacher}) . "</user>");
+   } elsif ($self->{teacher_inlay}) {
+      (delete $self->{teacher_inlay})->clear;
+   }
+
+   $self->update_cursor;
+}
+
 sub event_update_rules {
    my ($self, $rules) = @_;
 
    $self->{rules} = $rules;
 
    if ($self->{user}) {
+      # todo. gets drawn wrongly
+
       $self->{userpanel}[$_]->configure ($self->{app}, $self->{user}[$_], $rules)
          for COLOUR_BLACK, COLOUR_WHITE;
    }
@@ -691,6 +539,244 @@ sub event_update_rules {
    $self->{rules_inlay}->refresh;
 }
 
+### BOARD DISPLAY ###########################################################
+
+sub update_timers {
+   my ($self, $timers) = @_;
+
+   my $running = $self->{showmove} == @{$self->{path}} && !$self->{teacher};
+
+   for my $colour (COLOUR_BLACK, COLOUR_WHITE) {
+      my $t = $timers->[$colour];
+      $self->{userpanel}[$colour]->set_timer (
+            $running && $colour == $self->{whosemove} && $t->[0],
+            $t->[1] || $self->{rules}{time}
+                       + ($self->{rules}{timesys} == TIMESYS_BYO_YOMI
+                         && $self->{rules}{interval} * $self->{rules}{count}),
+            $t->[2]);
+   }
+}
+
+sub inject_set_gametime {
+   my ($self, $msg) = @_;
+
+   $self->{timers} = [
+      [$msg->{NOW}, $msg->{black_time}, $msg->{black_moves}],
+      [$msg->{NOW}, $msg->{white_time}, $msg->{white_moves}],
+   ];
+
+   $self->update_timers ($self->{timers})
+      if $self->{showmove} == @{$self->{path}};
+}
+
+sub update_cursor {
+   my ($self) = @_;
+
+   return unless $self->{cur_board};
+
+   if ($self->{rules}{ruleset} == RULESET_JAPANESE) {
+      if ($self->{curnode}{move} == 0) {
+         $self->{whosemove} = $self->{handicap} ? COLOUR_WHITE : COLOUR_BLACK;
+      } else {
+         $self->{whosemove} = 1 - $self->{cur_board}{last};
+      }
+   } else {
+      # Chinese, Aga, NZ all have manual placement
+      if ($self->{curnode}{move} < $self->{handicap}) {
+         $self->{whosemove} = COLOUR_BLACK;
+      } elsif ($self->{curnode}{move} == $self->{handicap}) {
+         $self->{whosemove} = $self->{handicap} ? COLOUR_WHITE : COLOUR_BLACK;
+      } else {
+         $self->{whosemove} = 1 - $self->{cur_board}{last};
+      }
+   }
+
+   my $running = $self->{showmove} == @{$self->{path}} && $self->is_active;
+
+   delete $self->{board_click};
+
+   if ($self->{teacher} eq $self->{app}{conn}) {
+      #TODO# # teaching mode not implemented
+      $self->{button_pass}->set (label => "Pass", sensitive => 1);
+      $self->{button_pass}->show;
+      $self->{button_undo}->hide;
+      $self->{button_resign}->hide;
+      $self->{board}->set (cursor => undef);
+
+   } elsif ($running && $self->{colour} != COLOUR_NONE) {
+      # during game
+      $self->{button_undo}->show;
+      $self->{button_resign}->show;
+
+      if ($self->{cur_board}{score}) {
+         # during scoring
+         $self->{button_pass}->set (label => "Done", sensitive => 1);
+         $self->{button_pass}->show;
+         $self->{board}->set (cursor => sub {
+            $_[0] & (MARK_B | MARK_W)
+               ? $_[0] ^ MARK_GRAYED
+               : $_[0];
+         });
+         $self->{board_click} = sub {
+            if ($_[0] == 255) {
+               $self->{button_pass}->sensitive (0);
+               $self->done;
+            } else {
+               $self->send (mark_dead =>
+                  channel => $self->{channel},
+                  x       => $_[0],
+                  y       => $_[1],
+                  dead    => !($self->{cur_board}{board}[$_[0]][$_[1]] & MARK_GRAYED),
+               );
+            }
+         };
+
+      } elsif ($self->{colour} == $self->{whosemove}) {
+         # normal move
+         $self->{button_pass}->set (label => "Pass", sensitive => 1);
+         $self->{button_pass}->show;
+         $self->{board}->set (cursor => sub {
+            $self->{cur_board}
+               && $self->{cur_board}->is_valid_move ($self->{colour}, $_[1], $_[2],
+                                         $self->{rules}{ruleset} == RULESET_NEW_ZEALAND)
+               ? $_[0] | MARK_GRAYED | ($self->{colour} == COLOUR_WHITE ? MARK_W : MARK_B)
+               : $_[0];
+         });
+         $self->{board_click} = sub {
+            return unless
+               $self->{cur_board}->is_valid_move ($self->{colour}, $_[0], $_[1],
+                                      $self->{rules}{ruleset} == RULESET_NEW_ZEALAND);
+            $self->send (game_move => channel => $self->{channel}, x => $_[0], y => $_[1]);
+            $self->{board}->set (cursor => undef);
+            delete $self->{board_click};
+            $self->{button_pass}->sensitive (0);
+         };
+      } else {
+         $self->{button_pass}->set (label => "Pass", sensitive => 0);
+         $self->{button_pass}->show;
+         $self->{board}->set (cursor => undef);
+      }
+   } else {
+      $self->{button_undo}->hide;
+      $self->{button_resign}->hide;
+      $self->{button_pass}->hide;
+      $self->{board}->set (cursor => undef);
+      #TODO# # implement coordinate-grabbing
+   }
+}
+
+sub update_board {
+   my ($self) = @_;
+
+   return unless $self->{path};
+
+   $self->{board_label}->set_text ("Move " . ($self->{showmove} - 1));
+
+   $self->{cur_board} = new KGS::Game::Board $self->{size};
+   $self->{cur_board}->interpret_path ([@{$self->{path}}[0 .. $self->{showmove} - 1]]);
+
+   $self->{userpanel}[$_]->set_captures ($self->{cur_board}{captures}[$_])
+      for COLOUR_WHITE, COLOUR_BLACK;
+
+   $self->{board}->set_board ($self->{cur_board});
+
+   if ($self->{cur_board}{score}) {
+      $self->{score_inlay} ||= $self->{chat}->new_inlay;
+      $self->{score_inlay}->clear;
+      $self->{score_inlay}->append_text ("\n<header>Scoring</header>"
+                                         . "\n<score>"
+                                         . "White: $self->{cur_board}{score}[COLOUR_WHITE], "
+                                         . "Black: $self->{cur_board}{score}[COLOUR_BLACK]"
+                                         . "</score>");
+   } elsif ($self->{score_inlay}) {
+      (delete $self->{score_inlay})->clear;
+   }
+
+   $self->update_cursor;
+
+   if ($self->{showmove} == @{$self->{path}}) {
+      $self->{timers} = [
+         [$self->{lastmove_time}, @{$self->{cur_board}{timer}[0]}],
+         [$self->{lastmove_time}, @{$self->{cur_board}{timer}[1]}],
+      ];
+      $self->update_timers ($self->{timers});
+   } else {
+      $self->update_timers ([
+         [0, @{$self->{cur_board}{timer}[0]}],
+         [0, @{$self->{cur_board}{timer}[1]}],
+      ]);
+   }
+
+}
+
+sub event_update_tree {
+   my ($self) = @_;
+
+   (delete $self->{undo_inlay})->clear
+      if $self->{undo_inlay};
+
+   $self->{path} = $self->get_path;
+
+   if ($self->{moveadj}) {
+      my $upper = $self->{moveadj}->upper;
+      my $pos = $self->{moveadj}->get_value;
+      my $move = scalar @{$self->{path}};
+
+      $self->{moveadj}->upper ($move);
+      
+      $self->{moveadj}->changed;
+      if ($pos == $upper) {
+         $self->{moveadj}->value ($move);
+         $self->{moveadj}->value_changed;
+      }
+   }
+}
+
+sub event_update_comments {
+   my ($self, $node, $comment, $newnode) = @_;
+   $self->SUPER::event_update_comments ($node, $comment, $newnode);
+
+   my $text;
+
+   $text .= "\n<header>Move <move>$node->{move}</move>, Node <node>$node->{id}</node></header>"
+      if $newnode;
+
+   for (split /\n/, $comment) {
+      $text .= "\n";
+      if (s/^([0-9a-zA-Z]+ \[[0-9dkp\?\-]+\])://) {
+         $text .= "<user>" . (util::toxml $1) . "</user>:";
+      }
+      
+      # coords only for 19x19 so far
+      $_ = util::toxml $_;
+      s{
+         (
+            \b
+            (?:[bw])?
+            [, ]{0,2}
+            [a-hj-t] # valid for upto 19x19
+            \s?
+            [1-9]?[0-9]
+            \b
+         )
+      }{
+         "<coord>$1</coord>";
+      }sgexi;
+
+      $text .= $_;
+   }
+
+   $self->{chat}->append_text ($text);
+}
+
+sub event_move {
+   my ($self, $pass) = @_;
+
+   sound::play 1, $pass ? "pass" : "move";
+}
+
+### GAMEPLAY EVENTS #########################################################
+
 sub event_resign_game {
    my ($self, $player) = @_;
 
@@ -698,17 +784,39 @@ sub event_resign_game {
    $self->{chat}->append_text ("\n<infoblock><header>Resign</header>"
                                . "\n<user>"
                                . (util::toxml $self->{user}[$msg->{player}]->as_string)
-                               . "</user> resigned.</infoblock>");
+                               . "</user> resigned."
+                               . "\n<user>"
+                               . (util::toxml $self->{user}[1 - $msg->{player}]->as_string)
+                               . "</user> wins the game."
+                               . "</infoblock>");
 }
 
-sub event_out_of_time {
+sub event_time_win {
    my ($self, $player) = @_;
 
    sound::play 3, "timewin";
    $self->{chat}->append_text ("\n<infoblock><header>Out of Time</header>"
+                               . "\n<user>"
+                               . (util::toxml $self->{user}[1 - $msg->{player}]->as_string)
+                               . "</user> ran out of time and lost."
                                . "\n<user>"
                                . (util::toxml $self->{user}[$msg->{player}]->as_string)
-                               . "</user> ran out of time and lost.</infoblock>");
+                               . "</user> wins the game."
+                               . "</infoblock>");
+}
+
+sub event_owner_left {
+   my ($self) = @_;
+
+   $self->{chat}->append_text ("\n<infoblock><header>Owner left</header>"
+                               . "\nThe owner of this game left.</infoblock>");
+}
+
+sub event_teacher_left {
+   my ($self) = @_;
+
+   $self->{chat}->append_text ("\n<infoblock><header>Teacher left</header>"
+                               . "\nThe teacher left the game.</infoblock>");
 }
 
 sub event_done {
@@ -740,18 +848,6 @@ sub inject_final_result {
                               );
 }
 
-sub inject_set_gametime {
-   my ($self, $msg) = @_;
-
-   $self->{timers} = [
-      [$msg->{NOW}, $msg->{black_time}, $msg->{black_moves}],
-      [$msg->{NOW}, $msg->{white_time}, $msg->{white_moves}],
-   ];
-
-   $self->update_timers ($self->{timers})
-      if $self->{showmove} == @{$self->{path}};
-}
-
 sub inject_req_undo {
    my ($self, $msg) = @_;
 
@@ -788,6 +884,8 @@ sub inject_new_game {
    delete $self->{cid};
 }
 
+### CHALLENGE HANDLING ######################################################
+
 sub draw_challenge {
    my ($self, $id) = @_;
 
@@ -802,7 +900,7 @@ sub draw_challenge {
 
    if (!$self->{channel}) {
       $inlay->append_text ("\nNotes: ");
-      $inlay->append_entry (\$info->{notes}, 20, "");
+      $inlay->append_widget (gtk::textentry \$info->{notes}, 20, "");
       $inlay->append_text ("\nGlobal Offer: ");
       $inlay->append_optionmenu (\$info->{flags},
          0 => "No",
@@ -865,7 +963,7 @@ sub draw_challenge {
       $inlay->append_optionmenu (\$info->{rules}{handicap}, map +($_, $_), 0..9);
 
       $inlay->append_text ("\nKomi: ");
-      $inlay->append_entry (\$info->{rules}{komi}, 5);
+      $inlay->append_widget (gtk::numentry \$info->{rules}{komi}, 5);
    }
 
    $inlay->append_text ("\nTimesys: ");
@@ -904,11 +1002,11 @@ sub draw_challenge {
    );
 
    $inlay->append_text ("\nMain Time: ");
-   $time = $inlay->append_entry (\$info->{rules}{time}, 5);
+   $time = $inlay->append_widget (gtk::timeentry \$info->{rules}{time}, 5);
    $inlay->append_text ("\nInterval: ");
-   $interval = $inlay->append_entry (\$info->{rules}{interval}, 3);
+   $interval = $inlay->append_widget (gtk::timeentry \$info->{rules}{interval}, 5);
    $inlay->append_text ("\nPeriods/Stones: ");
-   $count = $inlay->append_entry (\$info->{rules}{count}, 2);
+   $count = $inlay->append_widget (gtk::numentry \$info->{rules}{count}, 5);
 
    $inlay->append_text ("\n");
 
@@ -952,12 +1050,27 @@ sub draw_challenge {
    }
 }
 
-sub draw_users {
-   my ($self, $inlay) = @_;
+sub new_game_challenge {
+   my ($self) = @_;
 
-   for (sort keys %{$self->{users}}) {
-      $inlay->append_text ("  <user>" . $self->{users}{$_}->as_string . "</user>");
-   }
+   my $d = $self->{app}{defaults};
+
+   $self->{challenge}{""} = {
+      gametype => $d->{gametype},
+      flags    => 0,
+      notes    => $d->{stones},
+      rules    => {
+         ruleset  => $d->{ruleset},
+         size     => $d->{size},
+         timesys  => $d->{timesys},
+         time     => $d->{time},
+         interval => $d->{timesys} == TIMESYS_BYO_YOMI ? $d->{byo_time}    : $d->{can_time},
+         count    => $d->{timesys} == TIMESYS_BYO_YOMI ? $d->{byo_periods} : $d->{can_stones},
+      },
+
+      inlay => $self->{chat}->new_inlay,
+   };
+   $self->draw_challenge ("");
 }
 
 sub event_challenge {
@@ -973,8 +1086,8 @@ sub event_challenge {
    $self->{challenge}{$id} = $info;
    $self->{challenge}{$id}{inlay} = $self->{chat}->new_switchable_inlay (
       exists $self->{challenge}{""}
-         ? "Challenge from $opponent->{name}"
-         : "Challenge to $opponent->{name}",
+         ? "Challenge from " . $opponent->as_string
+         : "Challenge to " . $opponent->as_string,
       sub {
          $self->{challenge}{$id}{inlay} = $_[0];
          $self->draw_challenge ($id);
